# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule

try:
    # Python 2
    unicode
except NameError:
    # Python 3
    unicode = str

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()

__metaclass__ = type


class ModClusterModule(BaseJBossModule):

    def __init__(self, context=None):
        super(ModClusterModule, self).__init__(path='/subsystem=modcluster', context=context)

    def apply(self, modcluster=None, **kwargs):
        """
        Apply a modcluster subsystem operation:

        Example:

        modcluster:
          state: present

        """
        if not isinstance(modcluster, dict):
            raise ParameterError('%s provided to %s is not an allowable type' % (modcluster, self.__class__.__name__))

        changes = []

        state = self._get_param(modcluster, 'state')

        if state not in ['present', 'absent']:
            raise ParameterError('modcluster subsystem state is not one of [present|absent]')

        if state == 'present':
            changes += self.apply_present()
        elif state == 'absent':
            changes += self.apply_absent()

        return None if len(changes) < 1 else changes

    def apply_absent(self):
        try:
            self.cmd('%s:remove()' % self.path)
            return [{'subsystem': 'modcluster', 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_present(self):
        try:
            self.read_resource('%s' % self.path)
            return []
        except NotFoundError:
            self.cmd('%s:add()' % self.path)
            return [{'subsystem': 'modcluster', 'action': 'add'}]



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from abc import ABCMeta

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule

__metaclass__ = type


class SecurityModule(BaseJBossModule):
    __metaclass__ = ABCMeta

    def __init__(self, context=None):
        super(SecurityModule, self).__init__(path='/subsystem=security', context=context)

    def apply(self, security=None, **kwargs):
        """
        Apply an extension operation:

        Example:

        security:
          security-domain:
            - name: keycloak
              state: present
              authentication:
                name: classic
                login-modules:
                  - code: "org.keycloak.adapters.jboss.KeycloakLoginModule"
                    flag: required

        :param security: a single dict of an extension
        :param kwargs: any other args that may be useful
        :return: changed flag and a list of changes that have been applied to the security domain
        """

        security = self.unescape_keys(security)
        changes = []

        if security is None or 'security-domain' not in security or security['security-domain'] is None:
            raise ParameterError(
                '%s.apply(): No valid security configuration provided: %r' % (self.__class__.__name__, security))

        if not isinstance(security['security-domain'], list):
            raise ParameterError(
                '%s.apply(security.security-domain:%s): The security domain is not a valid list.' % (
                    self.__class__.__name__, type(security['security-domain'])))

        for domain in security['security-domain']:

            state = self._get_param(domain, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('security domain state is not one of [present|absent]')

            name = self._get_param(domain, 'name')

            if state == 'present':
                changes += self.apply_security_domain_present(**domain)
            elif state == 'absent':
                changes += self.apply_security_domain_absent(name)

        return None if len(changes) < 1 else changes

    def apply_security_domain_present(self, name=None, authentication=None, **kwargs):

        resource_path = '%s/security-domain=%s' % (self.path, name)
        try:
            self.read_resource_dmr(resource_path, True)
            # TODO sync authentication configuration
            return []
        except NotFoundError:
            # create the domain and add auth modules if present
            self.cmd('%s:add' % resource_path)
            if authentication is not None:
                auth_type = authentication.get('type', 'classic')
                modules = authentication.get('login-modules', [])
                self.cmd('%s/authentication=%s:add(login-modules=%s)' % (
                    # FIXME use convert_to_dmr_params
                    resource_path, auth_type, self.converts_to_dmr(modules)))
            return [{'security-domain': name, 'action': 'add'}]

    def apply_security_domain_absent(self, name):
        try:
            self.cmd('%s/security-domain=%s:remove' % (self.path, name))
            return [{'security-domain': name, 'action': 'delete'}]
        except NotFoundError:
            return []



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import copy

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule
from jyboss.logging import debug

__metaclass__ = type

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()

try:
    import simplejson as json
except ImportError:
    import json


class JGroupsModule(BaseJBossModule):
    JGROUPS_PARAMS = [
        'default-channel',
        'default-stack'
    ]

    def __init__(self, context=None):
        super(JGroupsModule, self).__init__(path='/subsystem=jgroups', context=context)
        self.CONFIG_SUBMODULE = {
            'stack': JGroupsStackModule(self.context),
            'channel': JGroupsChannelModule(self.context)
        }

    def apply(self, jgroups=None, **kwargs):

        jgroups = self.unescape_keys(jgroups)

        if not isinstance(jgroups, dict):
            raise ParameterError('%s provided to %s is not an allowable type' % (jgroups, self.__class__.__name__))

        changes = []

        state = self._get_param(jgroups, 'state')
        if state not in ['present', 'absent']:
            raise ParameterError('The jgroups state is not one of [present|absent]')

        # first need to deal with jgroups submodule itself
        if state == 'present':
            changes += self.apply_present(jgroups)

        elif state == 'absent':
            changes += self.apply_absent()

        return None if len(changes) < 1 else changes

    def apply_present(self, jgroups):
        changes = []

        try:
            self.read_resource(self.path, recursive=True)
        except NotFoundError:
            self.cmd('%s:add()' % self.path)
            changes.append({'subsystem': 'jgroups', 'action': 'add'})

        for key in jgroups:
            if key in self.CONFIG_SUBMODULE:
                handler = self.CONFIG_SUBMODULE[key]
                subchanges = handler.apply(jgroups[key])
                if len(subchanges) > 0:
                    changes.append({
                        key: subchanges
                    })
            elif key in JGroupsModule.JGROUPS_PARAMS:
                # very simple hack to set or unset value
                jgroups_dmr = self.read_resource_dmr(self.path)
                fc = dict(
                    (k, v) for (k, v) in iteritems(jgroups) if k in self.JGROUPS_PARAMS
                )
                changes += self._sync_attributes(parent_node=jgroups_dmr, parent_path=self.path, target_state=fc,
                                                 allowable_attributes=self.JGROUPS_PARAMS)
            elif key in ['state']:
                pass
            else:
                raise ParameterError('%s cannot handle configuration %s' % (self.__class__.__name__, key))

        return changes

    def apply_absent(self):
        changes = []

        try:
            self.read_resource_dmr(self.path)
            self.cmd('%s:remove()' % self.path)
            changes.append({'subsystem': 'jgroups', 'action': 'delete'})
        except NotFoundError:
            pass

        return changes


class JGroupsStackModule(BaseJBossModule):
    STACK_PARAMS = [
        'transport',
        'protocol'
    ]

    TRANSPORT_PARAMS = [
        'shared',
        'rack',
        'socket-binding'
    ]

    PROTOCOL_PARAMS = [
        'socket-binding',
        'datasource-jndi-name',
        'data-source',
        'frag-size'
    ]

    def __init__(self, context=None):
        super(JGroupsStackModule, self).__init__(path='/subsystem=jgroups/stack=%s', context=context)

    def apply(self, stack=None, **kwargs):

        stack = self._format_apply_param(stack)

        changes = []

        for st in stack:
            state = self._get_param(st, 'state')
            if state not in ['present', 'absent']:
                raise ParameterError('The stack state is not one of [present|absent]')

            if state == 'present':
                changes += self.apply_present(st)
            elif state == 'absent':
                changes += self.apply_absent(st)

        return changes

    def apply_absent(self, stack):

        name = self._get_param(stack, 'name')
        try:
            self.cmd((self.path + ':remove()') % name)
            return [{'stack': name, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_present(self, stack):
        name = self._get_param(stack, 'name')
        stack_path = self.path % name
        changes = []

        try:
            self.read_resource_dmr(stack_path, recursive=True)
        except NotFoundError:
            self.cmd('%s:add()' % stack_path)
            changes.append({'stack': name, 'action': 'add'})

        # now sync sub items
        for k, v in iteritems(stack):
            if k == 'transport':
                changes += self.sync_transport(name, v)
            elif k == 'protocol':
                changes += self.sync_protocols(name, v)
            elif k in ['name', 'state']:
                pass
            else:
                raise NotImplementedError('%s.apply(stack) does not understand how to apply %s sub configuration' % \
                                          (self.__class__.__name__, k))

        return changes

    def sync_transport(self, stack_name, transport):

        changes = []
        name = self._get_param(transport, 'type')
        transport_path = (self.path % stack_name) + ('/transport=%s' % name)

        # remove any transport that does not match the specified transport type in the new configuration
        tnames = self.ls((self.path % stack_name) + '/transport', silent=True)
        for tname in tnames:
            if tname not in ['TRANSPORT', name]:
                try:
                    self.cmd('%s:remove()' % transport_path)
                    changes.append({'transport': name, 'action': 'delete'})
                except NotFoundError:
                    pass

        # apply transport configuration
        try:
            transport_dmr = self.read_resource_dmr(transport_path)
            fc = dict(
                (k, v) for (k, v) in iteritems(transport) if k in self.TRANSPORT_PARAMS
            )
            a_changes = self._sync_attributes(parent_node=transport_dmr,
                                              parent_path=transport_path,
                                              target_state=fc,
                                              allowable_attributes=self.TRANSPORT_PARAMS)
            if len(a_changes) > 0:
                changes.append({'transport': name, 'action': 'update', 'changes': a_changes})
        except NotFoundError:
            tp = copy.deepcopy(transport)
            tp['type'] = name
            transport_params = self.convert_to_dmr_params(tp, self.TRANSPORT_PARAMS + ['type'])
            self.cmd('%s:add(%s)' % (transport_path, transport_params))
            changes.append({'transport': name, 'action': 'add', 'params': transport_params})

        return changes

    def sync_protocols(self, stack_name, protocols):
        """
        Synching protocols is a bit more complicated in jgroups as order of items is important. This method will check
        for ordered equality of the list of protocols and if equal will attempt to sync each item or if not equal will
        remove all and add them.
        :param stack_name: the name of the stack within the protocol
        :param protocols: the list of protocols to sync
        :return: any changes that have been applied in the process
        """
        changes = []

        if self._check_protocol_order(stack_name, protocols):
            # synch properties of each proto - can asume the order and quantity of protos is all good
            for new_proto in protocols:
                protocol_type = self._get_param(new_proto, 'type')
                old_proto = self.read_resource('{0}/protocol={1}'.format(self.path % stack_name, protocol_type))

                # FIXME sync all registered params

                # sync properties params
                # match all new to all old properties
                old_properties = old_proto.get('properties', {})
                for property in new_proto.get('properties', []):
                    p_k = property.get('name', None)
                    p_vn = property.get('value', None)
                    # first check is for a deliberate remove of a prop by setting new value to None
                    if p_vn is None:
                        if p_k in old_properties:
                            # remove it
                            self.cmd('{0}/protocol={1}/property={2}/:remove()' \
                                     .format(self.path % stack_name, protocol_type, p_k))
                            changes.append({
                                'protocol': protocol_type,
                                'property': p_k,
                                'action': 'delete'
                            })
                    # then we check if we need to compare both
                    elif p_k in old_properties:
                        # FIXME convert to comparable values
                        p_vo = old_properties.get(p_k)
                        # only a crappy hack atm
                        if isinstance(p_vo, dict) and p_vo.keys()[0] == 'EXPRESSION_VALUE':
                            p_vo = p_vo['EXPRESSION_VALUE']

                        if str(p_vo) != str(p_vn):
                            # update it
                            self.cmd('{0}/protocol={1}/property={2}/:add(value={3})' \
                                     .format(self.path % stack_name, protocol_type, p_k, p_vn))
                            changes.append({
                                'protocol': protocol_type,
                                'property': p_k,
                                'action': 'update',
                                'old_value': p_vo,
                                'new_value': p_vn
                            })
                    # lastly we may need to add if not exists
                    else:
                        # add it
                        self.cmd('{0}/protocol={1}/property={2}/:add(value={3})' \
                                 .format(self.path % stack_name, protocol_type, p_k, p_vn))
                        changes.append({
                            'protocol': protocol_type,
                            'property': p_k,
                            'action': 'add',
                            'new_value': p_vn
                        })

        else:
            # delete old protocols and add new set (we don't care about order here)
            old_protos = self.ls((self.path % stack_name) + '/protocol', silent=True)
            for old_proto in old_protos:
                self.cmd((self.path % stack_name) + ('/protocol=%s:remove()' % old_proto))
                changes.append({'protocol': old_proto, 'action': 'delete'})
            # add new set
            for new_proto in protocols:
                protocol_type = self._get_param(new_proto, 'type')
                proto_params = self.convert_to_dmr_params(new_proto, None)
                self.cmd('{0}/protocol={1}:add({2})'.format(self.path % stack_name, protocol_type, proto_params))
                properties = new_proto.get('properties', [])
                for new_prop in properties:
                    p_k = new_prop.get('name')
                    p_v = new_prop.get('value')
                    if p_v is not None:
                        self.cmd('{0}/protocol={1}/property={2}/:add(value={3})' \
                                 .format(self.path % stack_name, protocol_type, p_k, p_v))
                changes.append({'protocol': protocol_type, 'action': 'add', 'parameters': proto_params,
                                'properties': properties})

        return changes

    def _check_protocol_order(self, stack_name, protocols):
        old_stack = self.read_resource_dmr(self.path % stack_name)
        if old_stack.has('protocol') and str(old_stack.get('protocol').type) != 'UNDEFINED':
            old_protos = [str(p) for p in old_stack.get('protocol').keys()]
        else:
            old_protos = []

        new_protos = [p.get('type') for p in protocols]

        equal = len(old_protos) > 0 and len(old_protos) == len(new_protos)

        while len(old_protos) > 0 and equal:
            if old_protos.pop() != new_protos.pop():
                equal = False

        return equal


class JGroupsChannelModule(BaseJBossModule):
    CHANNEL_PARAMS = [
        'address',
        'address-as-uuid',
        'cluster',
        'discard-own-messages',
        'num-tasks-in-timer',
        'num-timer-threads',
        'received-bytes',
        'received-messages',
        'sent-bytes',
        'sent-messages',
        'stack',
        'stats-enabled',
        'version',
        'view'
    ]

    def __init__(self, context=None):
        super(JGroupsChannelModule, self).__init__(path='/subsystem=jgroups/channel=%s', context=context)

    def apply(self, channel=None, **kwargs):

        channel = self._format_apply_param(channel)

        changes = []

        for ch in channel:
            state = self._get_param(ch, 'state')
            if state not in ['present', 'absent']:
                raise ParameterError('The channel state is not one of [present|absent]')

            if state == 'present':
                changes += self.apply_present(ch)
            elif state == 'absent':
                changes += self.apply_absent(ch)

        return changes

    def apply_absent(self, stack):

        name = self._get_param(stack, 'name')
        try:
            self.cmd((self.path + ':remove()') % name)
            return [{'channel': name, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_present(self, channel):
        name = self._get_param(channel, 'name')
        channel_path = self.path % name
        changes = []

        try:
            channel_dmr = self.read_resource_dmr(channel_path, recursive=True)
            fc = dict(
                (k, v) for (k, v) in iteritems(channel) if k in self.CHANNEL_PARAMS)
            a_changes = self._sync_attributes(parent_node=channel_dmr,
                                              parent_path=channel_path,
                                              target_state=fc,
                                              allowable_attributes=self.CHANNEL_PARAMS)
            if len(a_changes) > 0:
                changes.append({'realm': name, 'action': 'update', 'changes': a_changes})

        except NotFoundError:
            channel_params = self.convert_to_dmr_params(channel, self.CHANNEL_PARAMS)
            self.cmd('%s:add(%s)' % (channel_path, channel_params))
            changes.append({'channel': name, 'action': 'add', 'params': channel_params})

        return changes



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from abc import ABCMeta

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule

__metaclass__ = type

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()


class KeycloakAdapterModule(BaseJBossModule):
    __metaclass__ = ABCMeta

    # TODO add all others from the sources
    KC_REALM_PARAMS = {
        'realm',
        'allow-any-hostname',
        'client-keystore-password',
        'cors-max-age',
        'realm-public-key',
        'truststore-password',
        'always-refresh-token',
        'client-keystore',
        'disable-trust-manager',
        'register-node-at-startup',
        'truststore',
        'auth-server-url-for-backend-requests',
        'connection-pool-size',
        'enable-cors',
        'register-node-period',
        'auth-server-url',
        'cors-allowed-headers',
        'expose-token',
        'ssl-required',
        'client-key-password',
        'cors-allowed-methods',
        'principal-attribute',
        'token-store'
    }

    KC_DEPLOYMENT_PARAMS = {
        'allow-any-hostname',
        'client-keystore-password',
        'disable-trust-manager',
        'realm-public-key',
        'token-store',
        'always-refresh-token',
        'client-keystore',
        'enable-basic-auth',
        'realm',
        'truststore-password',
        'auth-server-url-for-backend-requests',
        'connection-pool-size',
        'enable-cors',
        'register-node-at-startup',
        'truststore',
        'auth-server-url',
        'cors-allowed-headers',
        'expose-token',
        'register-node-period',
        'turn-off-change-session-id-on-login',
        'bearer-only',
        'cors-allowed-methods',
        'principal-attribute',
        'resource',
        'use-resource-role-mappings',
        'client-key-password',
        'cors-max-age',
        'public-client',
        'ssl-required'
    }

    def __init__(self, context=None):
        super(KeycloakAdapterModule, self).__init__(path='/subsystem=keycloak', context=context)

    def apply(self, keycloak_adapter=None, **kwargs):
        """
        Apply a keycloak adapter subsystem configuration:

        Example:
            keycloak_adapter:
              realm:
                - name: myrealm
                  state: present
                  realm_public_key: "MIIBIjANBgkq...A"
                  auth_server_url: 'https://login.server.com/auth'
                  ssl_required: none
                  enable_cors: false
                  principal_attribute: preferred_username
              secure_deployment:
                - name: myapp.war
                  state: present
                  resource: myrealm
                  credential:
                    type: secret
                    value: "5af5458f-0a96-4251-8f92-08ebcc3a8aa2"
                  disable_trust_manager: true
                  bearer_only: true
                  enable_basic_auth: true

        :param keycloak_adapter: the keycloak subsystem configuration to configure
        :param kwargs: any other args that may be useful
        :return: changed flag and a list of changes that have been applied to the security domain
        """

        keycloak = self.unescape_keys(keycloak_adapter)
        changes = []

        if keycloak is None or 'realm' not in keycloak or keycloak['realm'] is None:
            raise ParameterError(
                '%s.apply(): No valid keycloak subsystem configuration provided: %r' % (
                    self.__class__.__name__, keycloak))

        state = self._get_param(keycloak, 'state', 'present')

        if state not in ['present', 'absent']:
            raise ParameterError('security domain state is not one of [present|absent]')

        if state == 'present':
            changes += self.apply_keycloak_subsystem_present()

            if 'realm' in keycloak:
                changes += self.apply_realm(keycloak['realm'])

            if 'secure-deployment' in keycloak:
                changes += self.apply_secure_deployment(keycloak['secure-deployment'])

        elif state == 'absent':
            changes += self.apply_keycloak_subsystem_absent()

        # caller of apply will expect None if nothing changed, else a list of changes
        return None if len(changes) < 1 else changes

    def apply_keycloak_subsystem_present(self):
        try:
            self.read_resource_dmr(self.path)
            return []
        except NotFoundError:
            # create realm
            self.cmd('%s:add' % self.path)
            return [{'subsystem': 'keycloak', 'action': 'add'}]

    def apply_keycloak_subsystem_absent(self):
        changes = []
        try:
            self.cmd('%s:remove' % self.path)
            changes.append({'subsystem': 'keycloak', 'action': 'delete'})
        except NotFoundError:
            pass
        return changes

    def apply_realm(self, realm=None):
        # lets sync the realm configuration
        if isinstance(realm, dict):
            realms = [realm]
        elif isinstance(realm, list):
            realms = realm
        else:
            raise ParameterError(
                '%s.apply(keycloak.realm:%s): The keycloak subsystem configuration is not valid.' % (
                    self.__class__.__name__, type(realm)))

        changes = []

        for realm in realms:

            state = self._get_param(realm, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('The keycloak realm state is not one of [present|absent]')

            name = self._get_param(realm, 'name')

            if state == 'present':
                changes += self.apply_realm_present(realm)
            elif state == 'absent':
                changes += self.apply_realm_absent(name)

        return changes

    def apply_realm_present(self, realm=None):

        name = self._get_param(realm, 'name')
        realm_path = '%s/realm=%s' % (self.path, name)

        changes = []

        try:
            realm_dmr = self.read_resource_dmr(realm_path, True)
            fc = dict(
                (k, v) for (k, v) in iteritems(realm) if k in self.KC_REALM_PARAMS)
            a_changes = self._sync_attributes(parent_node=realm_dmr,
                                              parent_path=realm_path,
                                              target_state=fc,
                                              allowable_attributes=self.KC_REALM_PARAMS)
            if len(a_changes) > 0:
                changes.append({'realm': name, 'action': 'update', 'changes': a_changes})

        except NotFoundError:
            # create the domain and add auth modules if present
            realm_params = self.convert_to_dmr_params(realm, self.KC_REALM_PARAMS)

            self.cmd('%s:add(%s)' % (realm_path, realm_params))
            changes.append({'realm': name, 'action': 'add', 'params': realm_params})

        return changes

    def apply_realm_absent(self, name):
        try:
            self.cmd('%s/realm=%s:remove' % (self.path, name))
            # TODO also remove secure deployments with it
            return [{'keycloak-realm': name, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_secure_deployment(self, deployment):
        deployments = []
        if isinstance(deployment, dict):
            deployments.append(deployment)
        elif isinstance(deployment, list) or deployment is None:
            deployments += deployment
        else:
            raise ParameterError('Deployment is not valid: %s' % (type(deployment)))

        changes = []

        for deployment in deployments:
            state = self._get_param(deployment, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('The deployment state is not one of [present|absent]')

            name = self._get_param(deployment, 'name')

            if state == 'present':
                changes += self.apply_secure_deployment_present(deployment)
            elif state == 'absent':
                changes += self.apply_secure_deployment_absent(name)

        return changes

    def apply_secure_deployment_present(self, deployment):

        name = self._get_param(deployment, 'name')

        depl_path = '%s/secure-deployment=%s' % (self.path, name)

        changes = []

        try:
            depl_dmr = self.read_resource_dmr(depl_path, True)
            fc = dict(
                (k, v) for (k, v) in iteritems(deployment) if k in self.KC_REALM_PARAMS)
            a_changes = self._sync_attributes(parent_node=depl_dmr,
                                              parent_path=depl_path,
                                              target_state=fc,
                                              allowable_attributes=self.KC_REALM_PARAMS)
            if len(a_changes) > 0:
                changes.append({'secure-deployment': name, 'action': 'update', 'changes': a_changes})

            if 'credential' in deployment:
                changes += self._apply_credentials(name, deployment['credential'])

        except NotFoundError:
            # create the domain and add auth modules if present
            depl_params = self.convert_to_dmr_params(deployment, self.KC_DEPLOYMENT_PARAMS)
            self.cmd('%s:add(%s)' % (depl_path, depl_params))

            if 'credential' in deployment:
                changes += self._apply_credentials(name, deployment['credential'])

            changes.append({'secure-deployment': name, 'action': 'add', 'params': depl_params})

        return changes

    def _apply_credentials(self, deployment_name, credential):

        cred_type = self._get_param(credential, 'type')
        cred_value = self._get_param(credential, 'value')

        cred_path = '%s/secure-deployment=%s/credential=%s' % (self.path, deployment_name, cred_type)

        changes = []

        if cred_value is None:
            try:
                self.cmd('%s:remove()' % cred_path)
                changes.append({'deployment': deployment_name, 'credential': cred_type, 'action': 'delete'})
            except NotFoundError:
                pass
        else:
            try:
                cred = self.read_resource(cred_path)
                if 'value' in cred and cred['value'] != cred_value:
                    self.cmd(
                        '%s:write-attribute(name=value, value=%s)' % (cred_path, cred_value))
                    changes.append({'deployment': deployment_name, 'credential': cred_type, 'action': 'update'})

            except NotFoundError:
                self.cmd('%s:add(value=%s)' % (cred_path, cred_value))
                changes.append({'deployment': deployment_name, 'credential': cred_type, 'action': 'update'})

        return changes

    def apply_secure_deployment_absent(self, name):
        try:
            self.cmd('%s/secure-deployment=%s:remove' % (self.path, name))
            return [{'secure-deployment': name, 'action': 'delete'}]
        except NotFoundError:
            return []


class KeycloakServerModule(BaseJBossModule):
    __metaclass__ = ABCMeta

    KC_PARAMS = [
        'master-realm-name',
        'web-context',
        'scheduled-task-interval',
        'providers'
    ]

    def __init__(self, context=None):
        super(KeycloakServerModule, self).__init__(path='/subsystem=keycloak-server', context=context)
        self.submodules = {
            'spi': KeycloakServerSpiModule(context=context),
            'theme': KeycloakServerThemeModule(context=context)
        }

    def apply(self, keycloak_server=None, **kwargs):
        """
        Apply a keycloak adapter subsystem configuration:

        Example:
            keycloak_server:
              web_context: '/auth'

        :param keycloak_server: the keycloak subsystem configuration to apply
        :param kwargs: any other args that may be useful
        :return: changed flag and a list of changes that have been applied to the security domain
        """

        keycloak = self.unescape_keys(keycloak_server)

        if not isinstance(keycloak, dict):
            raise ParameterError('%s.apply(): No valid keycloak-server subsystem configuration provided: %r' % (
                self.__class__.__name__, keycloak))

        changes = []

        state = self._get_param(keycloak, 'state')
        if state not in ['present', 'absent']:
            raise ParameterError('The keycloak-server state is not one of [present|absent]')

        if state == 'present':
            changes += self.apply_present(keycloak)
        elif state == 'absent':
            changes += self.apply_absent()

        return changes if len(changes) > 0 else None

    def apply_present(self, keycloak):
        changes = []

        try:
            kc_dmr = self.read_resource_dmr(self.path, recursive=False)
            fc = dict((k, v) for (k, v) in iteritems(keycloak) if k in self.KC_PARAMS)
            a_changes = self._sync_attributes(parent_node=kc_dmr,
                                              parent_path=self.path,
                                              target_state=fc,
                                              allowable_attributes=self.KC_PARAMS)
            if len(a_changes) > 0:
                changes.append({'subsystem': 'keycloak-server', 'action': 'update', 'changes': a_changes})

        except NotFoundError:
            self.cmd('%s:add()' % self.path)
            changes.append({'subsystem': 'keycloak-server', 'action': 'add'})
            # TODO add all attributes to the subsystem as well

        # apply submodule configuration
        for key in keycloak:
            # first check if a submodule handler exists for the key element
            if key in self.submodules:
                handler = self.submodules[key]
                handler_changes = handler.apply(keycloak[key])
                if len(handler_changes) > 0:
                    changes += handler_changes
            elif key in ['state'] or key in self.KC_PARAMS:
                pass
            else:
                raise ParameterError('%s cannot handle configuration %s' % (self.__class__.__name__, key))

        return changes

    def apply_absent(self):
        changes = []

        try:
            self.read_resource_dmr(self.path)
            self.cmd('%s:remove()' % self.path)
            changes.append({'subsystem': 'keycloak-server', 'action': 'delete'})
        except NotFoundError:
            pass

        return changes


class KeycloakServerSpiModule(BaseJBossModule):
    __metaclass__ = ABCMeta

    SPI_PARAMS = [
        'default-provider'
    ]

    def __init__(self, context=None):
        super(KeycloakServerSpiModule, self).__init__(path='/subsystem=keycloak-server/spi=%s', context=context)
        self.submodules = {
            'providers': KeycloakServerSpiProviderModule(context=context)
        }

    def apply(self, spi=None, **kwargs):
        changes = []

        spi = self._format_apply_param(spi)

        for spi_module in spi:
            state = self._get_param(spi_module, 'state')
            if state == 'present':
                changes += self.apply_present(spi_module)
            elif state == 'absent':
                changes += self.apply_absent(spi_module)
            else:
                raise ParameterError('The spi state is not one of [present|absent]')

        return changes

    def apply_present(self, spi_module):
        name = self._get_param(spi_module, 'name')
        spi_path = self.path % name
        change = None

        try:
            spi_dmr = self.read_resource_dmr(spi_path, recursive=True)
            fc = dict((k, v) for (k, v) in iteritems(spi_module) if k in self.SPI_PARAMS)
            a_changes = self._sync_attributes(parent_node=spi_dmr,
                                              parent_path=spi_path,
                                              target_state=fc,
                                              allowable_attributes=self.SPI_PARAMS)
            if len(a_changes) > 0:
                change = {
                    'spi': name,
                    'action': 'update',
                    'changes': a_changes
                }
        except NotFoundError:
            spi_params = self.convert_to_dmr_params(spi_module, self.SPI_PARAMS)
            self.cmd('%s:add(%s)' % (spi_path, spi_params))
            change = {
                'spi': name,
                'action': 'add',
                'attributes': dict((k, v) for (k, v) in iteritems(spi_module) if k in self.SPI_PARAMS)
            }

        a_changes = self.submodules['providers'].apply(**spi_module)
        if len(a_changes) > 0:
            if change is None:
                change = {'spi': name, 'action': 'update'}
            # if other changes exist from parameter synching we add else we just
            if 'changes' in change:
                change['changes'] += a_changes
            else:
                change['changes'] = a_changes

        return [] if change is None else [change]

    def apply_absent(self, spi_module):
        name = self._get_param(spi_module, 'name')
        spi_path = self.path % name
        changes = []

        try:
            self.read_resource_dmr(spi_path)
            self.cmd('%s:remove()' % spi_path)
            changes.append({'spi': name, 'action': 'delete'})
        except NotFoundError:
            pass

        return changes


class KeycloakServerSpiProviderModule(BaseJBossModule):
    __metaclass__ = ABCMeta

    PROVIDER_PARAMS = [
        'enabled',
        'properties'
    ]

    def __init__(self, context=None):
        super(KeycloakServerSpiProviderModule, self).__init__(path='/subsystem=keycloak-server/spi=%s/provider=%s',
                                                              context=context)

    def apply(self, name=None, providers=None, **kwargs):
        changes = []

        providers = self._format_apply_param(providers)

        for provider in providers:
            state = self._get_param(provider, 'state')
            if state == 'present':
                changes += self.apply_present(name, provider)
            elif state == 'absent':
                changes += self.apply_absent(name, provider)
            else:
                raise ParameterError('The spi provider state is not one of [present|absent]')

        return changes

    def apply_present(self, spi_name, provider):
        name = self._get_param(provider, 'name')
        provider_path = self.path % (spi_name, name)
        changes = []

        try:
            provider_dmr = self.read_resource_dmr(provider_path, recursive=True)
            fc = dict((k, v) for (k, v) in iteritems(provider) if k in self.PROVIDER_PARAMS)
            a_changes = self._sync_attributes(parent_node=provider_dmr,
                                              parent_path=provider_path,
                                              target_state=fc,
                                              allowable_attributes=self.PROVIDER_PARAMS)
            if len(a_changes) > 0:
                changes.append({
                    'provider': name,
                    'action': 'update',
                    'changes': a_changes
                })

        except NotFoundError:
            provider_args = self.convert_to_dmr_params(provider, self.PROVIDER_PARAMS)
            self.cmd('%s:add(%s)' % (provider_path, provider_args))
            changes.append({
                'provider': name,
                'action': 'add',
                'attributes': dict((k, v) for (k, v) in iteritems(provider) if k in self.PROVIDER_PARAMS)
            })

        return changes

    def apply_absent(self, spi_name, provider):
        name = self._get_param(provider, 'name')
        provider_path = self.path % (spi_name, name)
        changes = []

        try:
            self.read_resource_dmr(provider_path)
            self.cmd('%s:remove()' % provider_path)
            changes.append({'provider': name, 'action': 'delete'})
        except NotFoundError:
            pass

        return changes


class KeycloakServerThemeModule(BaseJBossModule):
    __metaclass__ = ABCMeta

    THEME_PARAMS = [
        'cacheTemplates',
        'cacheThemes',
        'dir',
        'modules',
        'staticMaxAge',
        'welcomeTheme'
    ]

    def __init__(self, context=None):
        # this is really only a singular property and the name is hardcoded to defaults
        super(KeycloakServerThemeModule, self) \
            .__init__(path='/subsystem=keycloak-server/theme=defaults', context=context)

    def apply(self, theme=None, **kwargs):
        changes = []

        theme = self.unescape_keys(theme)

        state = self._get_param(theme, 'state')
        if state == 'present':
            changes += self.apply_present(theme)
        elif state == 'absent':
            changes += self.apply_absent()
        else:
            raise ParameterError('The theme state is not one of [present|absent]')

        return changes

    def apply_present(self, theme):
        change = None

        try:
            theme_dmr = self.read_resource_dmr(self.path, recursive=True)
            fc = dict((k, v) for (k, v) in iteritems(theme) if k in self.THEME_PARAMS)
            a_changes = self._sync_attributes(parent_node=theme_dmr,
                                              parent_path=self.path,
                                              target_state=fc,
                                              allowable_attributes=self.THEME_PARAMS)
            if len(a_changes) > 0:
                change = {
                    'theme': 'defaults',
                    'action': 'update',
                    'changes': a_changes
                }
        except NotFoundError:
            theme_params = self.convert_to_dmr_params(theme, self.THEME_PARAMS)
            self.cmd('%s:add(%s)' % (self.path, theme_params))
            change = {
                'theme': 'defaults',
                'action': 'add',
                'attributes': dict((k, v) for (k, v) in iteritems(theme) if k in self.THEME_PARAMS)
            }

        return [] if change is None else [change]

    def apply_absent(self):
        changes = []

        try:
            self.read_resource_dmr(self.path)
            self.cmd('%s:remove()' % self.path)
            changes.append({'theme': 'defaults', 'action': 'delete'})
        except NotFoundError:
            pass

        return changes



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

import sys
import os
import types
import re
from collections import OrderedDict

import datetime

from itertools import repeat, chain

try:
    # Python 2
    from itertools import imap
except ImportError:
    # Python 3
    imap = map

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()

try:
    # Python 2
    basestring
except NameError:
    # Python 3
    basestring = str

try:
    # Python 2
    unicode
except NameError:
    # Python 3
    unicode = str

# Python2 & 3 way to get NoneType
NoneType = type(None)

try:
    NUMBERTYPES = (int, long, float)
except NameError:
    # Python 3
    NUMBERTYPES = (int, float)

try:
    from collections import Sequence, Mapping
except ImportError:
    # python2.5
    Sequence = (list, tuple)
    Mapping = (dict,)

try:
    from collections.abc import KeysView

    SEQUENCETYPE = (Sequence, KeysView)
except:
    SEQUENCETYPE = Sequence

try:
    import json

    # Detect the python-json library which is incompatible
    # Look for simplejson if that's the case
    try:
        if not isinstance(json.loads, types.FunctionType) or not isinstance(json.dumps, types.FunctionType):
            raise ImportError
    except AttributeError:
        raise ImportError
except ImportError:
    try:
        import simplejson as json
    except ImportError:
        print(
            '\n{"msg": "Error: ansible requires the stdlib json or simplejson module, neither was found!", "failed": true}')
        sys.exit(1)
    except SyntaxError:
        print(
            '\n{"msg": "SyntaxError: probably due to installed simplejson being for a different python version", "failed": true}')
        sys.exit(1)

try:
    from ast import literal_eval
except ImportError:
    # a replacement for literal_eval that works with python 2.4. from:
    # https://mail.python.org/pipermail/python-list/2009-September/551880.html
    # which is essentially a cut/paste from an earlier (2.6) version of python's
    # ast.py
    from compiler import ast, parse


    def literal_eval(node_or_string):
        """
        Safely evaluate an expression node or a string containing a Python
        expression.  The string or node provided may only consist of the  following
        Python literal structures: strings, numbers, tuples, lists, dicts,  booleans,
        and None.
        """
        _safe_names = {'None': None, 'True': True, 'False': False}
        if isinstance(node_or_string, basestring):
            node_or_string = parse(node_or_string, mode='eval')
        if isinstance(node_or_string, ast.Expression):
            node_or_string = node_or_string.node

        def _convert(node):
            if isinstance(node, ast.Const) and isinstance(node.value, (basestring, int, float, long, complex)):
                return node.value
            elif isinstance(node, ast.Tuple):
                return tuple(map(_convert, node.nodes))
            elif isinstance(node, ast.List):
                return list(map(_convert, node.nodes))
            elif isinstance(node, ast.Dict):
                return dict((_convert(k), _convert(v)) for k, v in node.items())
            elif isinstance(node, ast.Name):
                if node.name in _safe_names:
                    return _safe_names[node.name]
            elif isinstance(node, ast.UnarySub):
                return -_convert(node.expr)
            raise ValueError('malformed string')

        return _convert(node_or_string)

__metaclass__ = type

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

_literal_eval = literal_eval


def get_exception():
    """Get the current exception.
    This code needs to work on Python 2.4 through 3.x, so we cannot use
    "except Exception, e:" (SyntaxError on Python 3.x) nor
    "except Exception as e:" (SyntaxError on Python 2.4-2.5).
    Instead we must use ::
        except Exception:
            e = get_exception()
    """
    return sys.exc_info()[1]


BOOLEANS_TRUE = ['yes', 'on', '1', 'true', 1, True]
BOOLEANS_FALSE = ['no', 'off', '0', 'false', 0, False]

# Internal global holding passed in params.  This is consulted in case
# multiple AnsibleModules are created.  Otherwise each AnsibleModule would
# attempt to read from stdin.  Other code should not use this directly as it
# is an internal implementation detail
_ANSIBLE_ARGS = None


def json_dict_unicode_to_bytes(d, encoding='utf-8'):
    """ Recursively convert dict keys and values to byte str

        Specialized for json return because this only handles, lists, tuples,
        and dict container types (the containers that the json module returns)
    """

    if isinstance(d, unicode):
        return d.encode(encoding)
    elif isinstance(d, dict):
        return OrderedDict(imap(json_dict_unicode_to_bytes, iteritems(d), repeat(encoding)))
    elif isinstance(d, list):
        return list(imap(json_dict_unicode_to_bytes, d, repeat(encoding)))
    elif isinstance(d, tuple):
        return tuple(imap(json_dict_unicode_to_bytes, d, repeat(encoding)))
    else:
        return d


def json_dict_bytes_to_unicode(d, encoding='utf-8'):
    """ Recursively convert dict keys and values to byte str

        Specialized for json return because this only handles, lists, tuples,
        and dict container types (the containers that the json module returns)
    """

    if isinstance(d, bytes):
        return unicode(d, encoding)
    elif isinstance(d, dict):
        return dict(imap(json_dict_bytes_to_unicode, iteritems(d), repeat(encoding)))
    elif isinstance(d, list):
        return list(imap(json_dict_bytes_to_unicode, d, repeat(encoding)))
    elif isinstance(d, tuple):
        return tuple(imap(json_dict_bytes_to_unicode, d, repeat(encoding)))
    else:
        return d


def return_values(obj):
    """ Return stringified values from datastructures. For use with removing
    sensitive values pre-jsonification."""
    if isinstance(obj, basestring):
        if obj:
            if isinstance(obj, bytes):
                yield obj
            else:
                # Unicode objects should all convert to utf-8
                # (still must deal with surrogateescape on python3)
                yield obj.encode('utf-8')
        return
    elif isinstance(obj, SEQUENCETYPE):
        for element in obj:
            for subelement in return_values(element):
                yield subelement
    elif isinstance(obj, Mapping):
        for element in obj.items():
            for subelement in return_values(element[1]):
                yield subelement
    elif isinstance(obj, (bool, NoneType)):
        # This must come before int because bools are also ints
        return
    elif isinstance(obj, NUMBERTYPES):
        yield str(obj)
    else:
        raise TypeError('Unknown parameter type: %s, %s' % (type(obj), obj))


def remove_values(value, no_log_strings):
    """ Remove strings in no_log_strings from value.  If value is a container
    type, then remove a lot more"""
    if isinstance(value, basestring):
        if isinstance(value, unicode):
            # This should work everywhere on python2. Need to check
            # surrogateescape on python3
            bytes_value = value.encode('utf-8')
            value_is_unicode = True
        else:
            bytes_value = value
            value_is_unicode = False
        if bytes_value in no_log_strings:
            return 'VALUE_SPECIFIED_IN_NO_LOG_PARAMETER'
        for omit_me in no_log_strings:
            bytes_value = bytes_value.replace(omit_me, '*' * 8)
        if value_is_unicode:
            value = unicode(bytes_value, 'utf-8', errors='replace')
        else:
            value = bytes_value
    elif isinstance(value, SEQUENCETYPE):
        return [remove_values(elem, no_log_strings) for elem in value]
    elif isinstance(value, Mapping):
        return dict((k, remove_values(v, no_log_strings)) for k, v in value.items())
    elif isinstance(value, tuple(chain(NUMBERTYPES, (bool, NoneType)))):
        stringy_value = str(value)
        if stringy_value in no_log_strings:
            return 'VALUE_SPECIFIED_IN_NO_LOG_PARAMETER'
        for omit_me in no_log_strings:
            if omit_me in stringy_value:
                return 'VALUE_SPECIFIED_IN_NO_LOG_PARAMETER'
    elif isinstance(value, datetime.datetime):
        value = value.isoformat()
    else:
        raise TypeError('Value of unknown type: %s, %s' % (type(value), value))
    return value


def _load_params():
    """ read the modules parameters and store them globally """
    global _ANSIBLE_ARGS

    buff = None
    if _ANSIBLE_ARGS is not None:
        buff = _ANSIBLE_ARGS
    else:
        if len(sys.argv) > 1:
            if os.path.isfile(sys.argv[1]):
                fd = open(sys.argv[1], 'rb')
                buff = fd.read()
                fd.close()
            else:
                buff = sys.argv[1]
        else:
            print(
                '\n{"msg": "Error: Module was not supplied with arguments in a JSON file. Unable to figure out what parameters were passed", "failed": true}')
            sys.exit(1)

        _ANSIBLE_ARGS = buff

    try:
        params = json.loads(buff.decode('utf-8'), object_pairs_hook=OrderedDict)
    except ValueError:
        # This helper used too early for fail_json to work.
        print(
            '\n{"msg": "Error: Module unable to decode valid JSON on stdin.  Unable to figure out what parameters were passed", "failed": true}')
        sys.exit(1)

    if PY2:
        params = json_dict_unicode_to_bytes(params)

    return params


class AnsibleModule(object):
    """simple implementation of the python module util shipped with ansible"""

    def __init__(self, argument_spec, bypass_checks=False, required_together=None, required_one_of=None,
                 required_if=None):
        self.argument_spec = argument_spec

        self._load_params()

        self._CHECK_ARGUMENT_TYPES_DISPATCHER = {
            'str': self._check_type_str,
            'list': self._check_type_list,
            'dict': self._check_type_dict,
            'bool': self._check_type_bool,
            'int': self._check_type_int,
            'float': self._check_type_float,
            'path': self._check_type_path,
            'raw': self._check_type_raw,
            'jsonarg': self._check_type_jsonarg,
        }
        if not bypass_checks:
            self._check_required_arguments()
            self._check_argument_types()
            self._check_argument_values()
            self._check_required_together(required_together)
            self._check_required_one_of(required_one_of)
            self._check_required_if(required_if)

        # Save parameter values that should never be logged
        self.no_log_values = set()

        # Use the argspec to determine which args are no_log
        for arg_name, arg_opts in self.argument_spec.items():
            if arg_opts.get('no_log', False):
                # Find the value for the no_log'd param
                no_log_object = self.params.get(arg_name, None)
                if no_log_object:
                    self.no_log_values.update(return_values(no_log_object))

        self.pretty_print = self.params.get('pretty', False)

    def fail_on_missing_params(self, required_params=None):
        """ This is for checking for required params when we can not check via argspec because we
        need more information than is simply given in the argspec.
        """
        if not required_params:
            return
        missing_params = []
        for required_param in required_params:
            if not self.params.get(required_param):
                missing_params.append(required_param)
        if missing_params:
            self.fail_json(msg="missing required arguments: %s" % ','.join(missing_params))

    def fail_json(self, **kwargs):
        """ return from the module, with an error message """
        assert 'msg' in kwargs, "implementation error -- msg to explain the error is required"
        kwargs['failed'] = True
        if 'invocation' not in kwargs:
            kwargs['invocation'] = {'module_args': self.params, 'interpreter': sys.executable}
        kwargs = remove_values(kwargs, self.no_log_values)
        # REVIEW self.do_cleanup_files()
        print('\n%s' % self.jsonify(kwargs))
        sys.exit(1)

    def exit_json(self, **kwargs):
        """ return from the module, without error """
        # REVIEW self.add_path_info(kwargs)
        if not 'changed' in kwargs:
            kwargs['changed'] = False
        if 'invocation' not in kwargs:
            kwargs['invocation'] = {'module_args': self.params, 'interpreter': sys.executable}
        kwargs = remove_values(kwargs, self.no_log_values)
        # REVIEW self.do_cleanup_files()
        print('\n%s' % self.jsonify(kwargs))
        sys.exit(0)

    def jsonify(self, data):
        for encoding in ("utf-8", "latin-1"):
            try:
                pretty = dict()
                if self.pretty_print:
                    pretty['indent'] = 4
                return json.dumps(data, encoding=encoding, **pretty)
            # Old systems using old simplejson module does not support encoding keyword.
            except TypeError:
                try:
                    new_data = json_dict_bytes_to_unicode(data, encoding=encoding)
                except UnicodeDecodeError:
                    continue
                return json.dumps(new_data)
            except UnicodeDecodeError:
                continue
        self.fail_json(msg='Invalid unicode encoding encountered')

    def boolean(self, arg):
        """ return a bool for the arg """
        if arg is None or type(arg) == bool:
            return arg
        if isinstance(arg, basestring):
            arg = arg.lower()
        if arg in BOOLEANS_TRUE:
            return True
        elif arg in BOOLEANS_FALSE:
            return False
        else:
            self.fail_json(msg='Boolean %s not in either boolean list' % arg)

    @staticmethod
    def safe_eval(str, locals=None, include_exceptions=False):

        # do not allow method calls to modules
        if not isinstance(str, basestring):
            # already templated to a datastructure, perhaps?
            if include_exceptions:
                return (str, None)
            return str
        if re.search(r'\w\.\w+\(', str):
            if include_exceptions:
                return (str, None)
            return str
        # do not allow imports
        if re.search(r'import \w+', str):
            if include_exceptions:
                return (str, None)
            return str
        try:
            result = literal_eval(str)
            if include_exceptions:
                return (result, None)
            else:
                return result
        except Exception:
            e = get_exception()
            if include_exceptions:
                return (str, e)
            return str

    @staticmethod
    def _check_type_str(value):
        if isinstance(value, basestring):
            return value
        # Note: This could throw a unicode error if value's __str__() method
        # returns non-ascii.  Have to port utils.to_bytes() if that happens
        return str(value)

    @staticmethod
    def _check_type_list(value):
        if isinstance(value, list):
            return value

        if isinstance(value, basestring):
            return value.split(",")
        elif isinstance(value, int) or isinstance(value, float):
            return [str(value)]

        raise TypeError('%s cannot be converted to a list' % type(value))

    def _check_type_dict(self, value):
        if isinstance(value, dict):
            return value

        if isinstance(value, basestring):
            if value.startswith("{"):
                try:
                    return json.loads(value)
                except:
                    (result, exc) = self.safe_eval(value, dict(), include_exceptions=True)
                    if exc is not None:
                        raise TypeError('unable to evaluate string as dictionary')
                    return result
            elif '=' in value:
                fields = []
                field_buffer = []
                in_quote = False
                in_escape = False
                for c in value.strip():
                    if in_escape:
                        field_buffer.append(c)
                        in_escape = False
                    elif c == '\\':
                        in_escape = True
                    elif not in_quote and c in ('\'', '"'):
                        in_quote = c
                    elif in_quote and in_quote == c:
                        in_quote = False
                    elif not in_quote and c in (',', ' '):
                        field = ''.join(field_buffer)
                        if field:
                            fields.append(field)
                        field_buffer = []
                    else:
                        field_buffer.append(c)

                field = ''.join(field_buffer)
                if field:
                    fields.append(field)
                return dict(x.split("=", 1) for x in fields)
            else:
                raise TypeError("dictionary requested, could not parse JSON or key=value")

        raise TypeError('%s cannot be converted to a dict' % type(value))

    def _check_type_bool(self, value):
        if isinstance(value, bool):
            return value

        if isinstance(value, basestring) or isinstance(value, int):
            return self.boolean(value)

        raise TypeError('%s cannot be converted to a bool' % type(value))

    @staticmethod
    def _check_type_int(value):
        if isinstance(value, int):
            return value

        if isinstance(value, basestring):
            return int(value)

        raise TypeError('%s cannot be converted to an int' % type(value))

    @staticmethod
    def _check_type_float(value):
        if isinstance(value, float):
            return value

        if isinstance(value, basestring):
            return float(value)

        raise TypeError('%s cannot be converted to a float' % type(value))

    def _check_type_path(self, value):
        value = self._check_type_str(value)
        return os.path.expanduser(os.path.expandvars(value))

    @staticmethod
    def _check_type_jsonarg(value):
        # Return a jsonified string.  Sometimes the controller turns a json
        # string into a dict/list so transform it back into json here
        if isinstance(value, (unicode, bytes)):
            return value.strip()
        else:
            if isinstance(value(list, tuple, dict)):
                return json.dumps(value)
        raise TypeError('%s cannot be converted to a json string' % type(value))

    @staticmethod
    def _check_type_raw(value):
        return value

    def _check_required_arguments(self):
        """ ensure all required arguments are present """
        missing = []
        for (k, v) in self.argument_spec.items():
            required = v.get('required', False)
            if required and k not in self.params:
                missing.append(k)
        if len(missing) > 0:
            self.fail_json(msg="missing required arguments: %s" % ",".join(missing))

    def _check_argument_types(self):
        """ ensure all arguments have the requested type """
        for (k, v) in self.argument_spec.items():
            wanted = v.get('type', None)

            # only if we have a wanted field we rampage and fiddle with the value
            # otherwise it would convert items we want to morph (e.g. param can be a list or single value)
            if wanted is None or k not in self.params or self.params[k] is None:
                continue

            value = self.params[k]

            try:
                type_checker = self._CHECK_ARGUMENT_TYPES_DISPATCHER[wanted]
            except KeyError:
                self.fail_json(msg="implementation error: unknown type %s requested for %s" % (wanted, k))
            try:
                self.params[k] = type_checker(value)
            except (TypeError, ValueError):
                self.fail_json(
                    msg="argument %s is of type %s and we were unable to convert to %s" % (k, type(value), wanted))

    def _check_argument_values(self):
        ''' ensure all arguments have the requested values, and there are no stray arguments '''
        for (k, v) in self.argument_spec.items():
            choices = v.get('choices', None)
            if choices is None:
                continue
            if isinstance(choices, SEQUENCETYPE):
                if k in self.params:
                    if self.params[k] not in choices:
                        choices_str = ",".join([str(c) for c in choices])
                        msg = "value of %s must be one of: %s, got: %s" % (k, choices_str, self.params[k])
                        self.fail_json(msg=msg)
            else:
                self.fail_json(msg="internal error: choices for argument %s are not iterable: %s" % (k, choices))

    def _count_terms(self, check):
        count = 0
        for term in check:
            if term in self.params:
                count += 1
        return count

    def _check_required_if(self, spec):
        """ ensure that parameters which conditionally required are present """
        if spec is None:
            return
        for (key, val, requirements) in spec:
            missing = []
            if key in self.params and self.params[key] == val:
                for check in requirements:
                    count = self._count_terms((check,))
                    if count == 0:
                        missing.append(check)
            if len(missing) > 0:
                self.fail_json(msg="%s is %s but the following are missing: %s" % (key, val, ','.join(missing)))

    def _check_required_together(self, spec):
        if spec is None:
            return
        for check in spec:
            counts = [self._count_terms([field]) for field in check]
            non_zero = [c for c in counts if c > 0]
            if len(non_zero) > 0:
                if 0 in counts:
                    self.fail_json(msg="parameters are required together: %s" % (check,))

    def _check_required_one_of(self, spec):
        if spec is None:
            return
        for check in spec:
            count = self._count_terms(check)
            if count == 0:
                self.fail_json(msg="one of the following is required: %s" % ','.join(check))

    def _check_mutually_exclusive(self, spec):
        if spec is None:
            return
        for check in spec:
            count = self._count_terms(check)
            if count > 1:
                self.fail_json(msg="parameters are mutually exclusive: %s" % (check,))

    def _load_params(self):
        ''' read the input and set the params attribute.

        This method is for backwards compatibility.  The guts of the function
        were moved out in 2.1 so that custom modules could read the parameters.
        '''
        # debug overrides to read args from file or cmdline
        self.params = _load_params()



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from abc import ABCMeta

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule
from jyboss.logging import debug

__metaclass__ = type

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()

try:
    import simplejson as json
except ImportError:
    import json


class InfinispanModule(BaseJBossModule):
    def __init__(self, context=None):
        super(InfinispanModule, self).__init__(path='/subsystem=infinispan', context=context)
        self.submodules = {
            'cache-container': CacheContainerModule(self.context)
        }

    def apply(self, infinispan=None, **kwargs):

        infinispan = self.unescape_keys(infinispan)

        if not isinstance(infinispan, dict):
            raise ParameterError('%s.apply(): No valid infinispan subsystem configuration provided: %r' % (
                self.__class__.__name__, infinispan))

        changes = []

        state = self._get_param(infinispan, 'state')
        if state not in ['present', 'absent']:
            raise ParameterError('The infinispan state is not one of [present|absent]')

        if state == 'present':
            changes += self.apply_present(infinispan)
        elif state == 'absent':
            changes += self.apply_absent()

        return changes if len(changes) > 0 else None

    def apply_present(self, infinispan):
        changes = []

        try:
            self.read_resource(self.path, recursive=False)
        except NotFoundError:
            self.cmd('%s:add()' % self.path)
            changes.append({'subsystem': 'infinispan', 'action': 'add'})

        for key in infinispan:
            # first check if a submodule handler exists for the key element
            if key in self.submodules:
                handler = self.submodules[key]
                handler_changes = handler.apply(infinispan[key])
                if len(handler_changes) > 0:
                    changes += handler_changes
            elif key in ['state']:
                pass
            else:
                raise ParameterError('%s cannot handle configuration %s' % (self.__class__.__name__, key))

        return changes

    def apply_absent(self):
        changes = []

        try:
            self.read_resource_dmr(self.path)
            self.cmd('%s:remove()' % self.path)
            changes.append({'subsystem': 'infinispan', 'action': 'delete'})
        except NotFoundError:
            pass

        return changes


class CacheContainerModule(BaseJBossModule):
    CONTAINER_PARAMS = [
        'default',
        'aliases',
        'jndi-name',
        'default-cache',
        'module',
        'statistics-enabled'
    ]

    def __init__(self, context=None):
        super(CacheContainerModule, self).__init__(path='/subsystem=infinispan/cache-container=%s',
                                                   context=context)
        self.submodules = {
            'transport': TransportModule(self.context),
            'caches': CacheResolverDelegate(self.context)
        }

    def apply(self, containers=None, **kwargs):

        containers = self._format_apply_param(containers)

        changes = []

        for container in containers:
            state = self._get_param(container, 'state')
            if state not in ['present', 'absent']:
                raise ParameterError('The container state is not one of [present|absent]')

            if state == 'present':
                changes += self.apply_present(container)
            elif state == 'absent':
                changes += self.apply_absent(container)

        return changes

    def apply_absent(self, stack):

        name = self._get_param(stack, 'name')
        try:
            self.cmd((self.path + ':remove()') % name)
            return [{'cache-container': name, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_present(self, container):
        container_name = self._get_param(container, 'name')
        container_path = self.path % container_name

        change = {
            'cache-container': container_name,
            'action': None,
            'changes': []
        }

        try:
            container_dmr = self.read_resource_dmr(container_path, recursive=True)
            fc = dict(
                (k, v) for (k, v) in iteritems(container) if k in self.CONTAINER_PARAMS)
            update_changes = self._sync_attributes(parent_node=container_dmr,
                                                   parent_path=container_path,
                                                   target_state=fc,
                                                   allowable_attributes=self.CONTAINER_PARAMS)
            if len(update_changes) > 0:
                change['action'] = 'update'
                change['changes'] += update_changes

        except NotFoundError:
            container_params = self.convert_to_dmr_params(container, self.CONTAINER_PARAMS)
            self.cmd('%s:add(%s)' % (container_path, container_params))
            change['action'] = 'add'
            if len(container_params) > 0:
                change['params'] = container_params

        # now apply all sub component handlers
        for key in container:
            if key in self.submodules:
                handler = self.submodules[key]
                modchanges = handler.apply(container_name, container[key])
                if len(modchanges) > 0:
                    if change['action'] is None:
                        change['action'] = 'update'
                    change['changes'] += modchanges
            elif key in ['state', 'name'] + CacheContainerModule.CONTAINER_PARAMS:
                pass  # not interested in these ones as they would have been processed above already
            else:
                raise ParameterError('%s cannot handle configuration %s' % (self.__class__.__name__, key))

        if len(change['changes']) < 1:
            change.pop('changes')

        return [] if change['action'] is None else [change]


class TransportModule(BaseJBossModule):
    TRANSPORT_PARAMS = [
        'channel',
        'stack',
        'cluster',
        'lock-timeout'
    ]

    def __init__(self, context=None):
        super(TransportModule, self).__init__(
            path='/subsystem=infinispan/cache-container=%s/transport=TRANSPORT',
            context=context)

    def apply(self, container_name, transport=None, **kwargs):

        changes = []

        state = self._get_param(transport, 'state')
        if state not in ['present', 'absent']:
            raise ParameterError('The container transport state is not one of [present|absent]')

        if state == 'present':
            changes += self.apply_present(container_name, transport)
        elif state == 'absent':
            changes += self.apply_absent(transport)

        return changes

    def apply_absent(self, container_name):

        try:
            self.cmd((self.path + ':remove()') % container_name)
            return [{'transport': 'TRANSPORT', 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_present(self, container_name, transport):
        transport_path = self.path % container_name
        change = {
            'transport': 'TRANSPORT',
            'action': None,
            'changes': []
        }

        try:
            transport_dmr = self.read_resource_dmr(transport_path, recursive=True)
            fc = dict(
                (k, v) for (k, v) in iteritems(transport) if k in self.TRANSPORT_PARAMS)
            a_changes = self._sync_attributes(parent_node=transport_dmr,
                                              parent_path=transport_path,
                                              target_state=fc,
                                              allowable_attributes=self.TRANSPORT_PARAMS)
            if len(a_changes) > 0:
                change['action'] = 'update'
                change['changes'] += a_changes

        except NotFoundError:
            transport_params = self.convert_to_dmr_params(transport, self.TRANSPORT_PARAMS)
            self.cmd('%s:add(%s)' % (transport_path, transport_params))
            change['action'] = 'add'
            if len(transport_params) > 0:
                change['params'] = transport_params

        if len(change['changes']) < 1:
            change.pop('changes')

        return [] if change['action'] is None else [change]


class CacheResolverDelegate(object):
    def __init__(self, context=None):
        self.submodules = {
            'local-cache': LocalCache(context),
            'replicated-cache': ReplicatedCache(context),
            'invalidation-cache': InvalidationCache(context),
            'distributed-cache': DistributedCache(context)
        }

    def apply(self, container_name, caches):

        changes = []

        for cache in caches:
            if 'type' not in cache:
                raise ParameterError('%s: cache type was not provided' % self.__class__.__name__)
            cache_type = cache['type']
            if cache_type in self.submodules:
                changes += self.submodules[cache_type].apply(container_name, cache)
            else:
                raise ParameterError(
                    '%s does not have a cache handler for type %s' % (self.__class__.__name__, cache_type))

        return changes


class Cache(BaseJBossModule):
    __metaclass__ = ABCMeta

    def __init__(self, context=None):
        super(Cache, self).__init__(path='/subsystem=infinispan/cache-container=%s/%s=%s', context=context)
        self.cache_params = [
            'jndi-name',
            'module',
            'statistics-enabled'
        ]

        self.cache_modules = {
            'locking': Component(component_type='locking',
                                 component_params=[
                                     'isolation',
                                     'striping',
                                     'acquire-timeout',
                                     'concurrency-level'
                                 ],
                                 context=context),
            'transaction': Component(component_type='transaction',
                                     component_params=[
                                         'mode',
                                         'stop-timeout',
                                         'locking'
                                     ],
                                     context=context),
            'eviction': Component(component_type='eviction',
                                  component_params=[
                                      'strategy',
                                      'max-entries'
                                  ],
                                  context=context),
            'expiration': Component(component_type='expiration',
                                    component_params=[
                                        'max-idle',
                                        'lifespan',
                                        'interval'
                                    ],
                                    context=context)
            # FIXME implement store cache submodules
            # store
            # file-store
            # string-keyed-jdbc-store
            # binary-keyed-jdbc-store
            # mixed-keyed-jdbc-store
            # remote-store
        }

    def apply(self, container_name, cache=None, **kwargs):
        changes = []
        state = self._get_param(cache, 'state')
        if state not in ['present', 'absent']:
            raise ParameterError('The cache state is not one of [present|absent]')

        if state == 'present':
            changes += self.apply_present(container_name, cache)
        elif state == 'absent':
            changes += self.apply_absent(container_name, cache)

        return changes

    def apply_absent(self, container_name, cache):
        cache_name = self._get_param(cache, 'name')
        cache_type = self._get_param(cache, 'type')
        try:
            self.cmd((self.path + ':remove()') % (container_name, cache_type, cache_name))
            return [{cache_type: cache_name, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_present(self, container_name, cache):
        cache_name = self._get_param(cache, 'name')
        cache_type = self._get_param(cache, 'type')
        cache_path = self.path % (container_name, cache_type, cache_name)

        change = {
            cache_type: cache_name,
            'action': None,
            'changes': []
        }

        try:
            cache_dmr = self.read_resource_dmr(cache_path, recursive=True)
            fc = dict(
                (k, v) for (k, v) in iteritems(cache) if k in self.cache_params)
            attr_changes = self._sync_attributes(parent_node=cache_dmr,
                                                 parent_path=cache_path,
                                                 target_state=fc,
                                                 allowable_attributes=self.cache_params)
            if len(attr_changes) > 0:
                change['action'] = 'update'
                change['changes'] += attr_changes

        except NotFoundError:
            cache_params = self.convert_to_dmr_params(cache, self.cache_params)
            self.cmd('%s:add(%s)' % (cache_path, cache_params))
            change['action'] = 'add'
            if len(cache_params) > 0:
                change['params'] = cache_params

        # process all submodules
        for key in cache:
            if key in self.cache_modules:
                modchanges = self.cache_modules[key].apply(container_name=container_name, cache_type=cache_type,
                                                           cache_name=cache_name, config=cache[key])
                if len(modchanges) > 0:
                    change[key] = modchanges

            elif key in self.cache_params or key in ['state', 'name', 'type']:
                pass
            else:
                raise ParameterError(
                    '%s does not have a component handler for %s' % (self.__class__.__name__, key))

        if len(change['changes']) < 1:
            change.pop('changes')

        return [] if change['action'] is None else [change]


class ClusteredCache(Cache):
    __metaclass__ = ABCMeta

    def __init__(self, context=None):
        super(ClusteredCache, self).__init__(context=context)
        self.cache_params += [
            'mode',
            'queue-size',
            'queue-flush-interval',
            'remote-timeout'
        ]


class SharedStateCache(ClusteredCache):
    __metaclass__ = ABCMeta

    def __init__(self, context=None):
        super(SharedStateCache, self).__init__(context=context)
        self.cache_modules['partition-handling'] = Component(component_type='partition-handling',
                                                             component_params=['enabled'],
                                                             context=context)
        self.cache_modules['state-transfer'] = Component(component_type='state-transfer',
                                                         component_params=['timeout', 'chunk-size'],
                                                         context=context)
        # FIXME implement backup cache submodules
        # backups
        # backup-for


class LocalCache(Cache):
    def __init__(self, context=None):
        super(LocalCache, self).__init__(context=context)


class InvalidationCache(ClusteredCache):
    def __init__(self, context=None):
        super(InvalidationCache, self).__init__(context=context)


class ReplicatedCache(SharedStateCache):
    def __init__(self, context=None):
        super(ReplicatedCache, self).__init__(context=context)


class DistributedCache(SharedStateCache):
    def __init__(self, context=None):
        super(DistributedCache, self).__init__(context=context)
        self.cache_params += [
            'owners',
            'segments',
            'l1-lifespan',
            'capacity-factor',
            'consistent-hash-strategy'
        ]


class Component(BaseJBossModule):
    def __init__(self, component_type, component_params=None, context=None):
        super(Component, self).__init__(path='/subsystem=infinispan/cache-container=%s/%s=%s/component=%s',
                                        context=context)
        self.component_type = component_type
        self.component_params = component_params

    def apply(self, container_name, cache_type, cache_name, config=None, **kwargs):
        if config is None:
            return self.apply_absent(container_name, cache_type, cache_name)
        else:
            return self.apply_present(container_name, cache_type, cache_name, config)

    def apply_absent(self, container_name, cache_type, cache_name):
        try:
            self.cmd((self.path + ':remove()') % (container_name, cache_type, cache_name, self.component_type))
            return [{'component': self.component_type, 'action': 'delete'}]
        except NotFoundError:
            return []

    def apply_present(self, container_name, cache_type, cache_name, config):
        config_path = self.path % (container_name, cache_type, cache_name, self.component_type)
        change = {
            'component': self.component_type,
            'action': None,
            'changes': []
        }

        try:
            config_dmr = self.read_resource_dmr(config_path, recursive=True)
            fc = dict(
                (k, v) for (k, v) in iteritems(config) if k in self.component_params)
            attr_changes = self._sync_attributes(parent_node=config_dmr,
                                                 parent_path=config_path,
                                                 target_state=fc,
                                                 allowable_attributes=self.component_params)
            if len(attr_changes) > 0:
                change['action'] = 'update'
                change['changes'] += attr_changes

        except NotFoundError:
            config_params = self.convert_to_dmr_params(config, self.component_params)
            self.cmd('%s:add(%s)' % (config_path, config_params))
            change['action'] = 'add'
            if len(config_params) > 0:
                change['params'] = config_params

        if len(change['changes']) < 1:
            change.pop('changes')

        return [] if change['action'] is None else [change]



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

# care command handlers
from .core import escape_keys, unescape_keys, expression_deserializer
from .core import ReloadCommandHandler, CliCmdHandler
from .core import ChangeObservable
from .core import CommandHandler as _BasicCommandHandler

# management specific handlers
from .undertow import UndertowModule
from .undertow import UndertowFilterRefModule
from .undertow import UndertowFilterModule
from .undertow import UndertowHttpListenerModule
from .undertow import UndertowAjpListenerModule

from .extension import ExtensionModule
from .ee import EEModule
from .weld import WeldModule
from .modcluster import  ModClusterModule
from .datasources import DatasourcesModule
from .module import ModuleModule
from .security import SecurityModule
from .keycloak import KeycloakAdapterModule
from .keycloak import KeycloakServerModule
from .deployment import DeploymentModule
from .jgroups import JGroupsModule
from .infinispan import InfinispanModule
from .interface import InterfaceModule
from .binding import SocketBindingModule

__metaclass__ = type

try:
    # Python 2
    unicode
except NameError:
    # Python 3
    unicode = str

"""
This package contains command modules to configure jboss subsystems.
"""

_base = _BasicCommandHandler()
# create shell alias for core functions
cd = _base.cd
ls = _base.ls
cmd = _base.cmd

batch = _base



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ContextError, CommandError

try:
    from java.lang import System, IllegalStateException, IllegalArgumentException
    from java.io import IOException
    from java.net import URI, URISyntaxException
except ImportError as jpe:
    raise ContextError('Java packages are not available, please run this module with jython.', jpe)

try:
    # @formatter:off
    # noinspection PyUnresolvedReferences
    from org.jboss.as.cli import CliInitializationException, CommandContext, CommandContextFactory, CommandFormatException, CommandLineException
    from org.jboss.as.cli.impl import CommandContextConfiguration
    from org.jboss.dmr import ModelNode
    # @formatter:on
except ImportError as jbe:
    raise ContextError(
        'The jboss client library is not present on the python path. Please configure the context classpath (se jyboss documentation).',
        jbe)

__author__ = 'Niels Bertram <nielsbne@gmail.com>'
__version__ = '0.2.8'

class Cli(object):
    """
    a replacement for the jboss scripting cli, which does not support embedded mode and has a few
    other quirks including a private constructor
    """

    def __init__(self):
        self.ctx = None

    @staticmethod
    def instance():
        return Cli()

    def is_silent(self):
        return False if self.ctx is None else self.ctx.isSilent()

    def set_silent(self, flag=True):
        if self.ctx is not None:
            self.ctx.setSilent(flag)

    def is_connected(self):
        return False if self.ctx is None else not self.ctx.isTerminated()

    def check_already_connected(self):
        if self.ctx is not None:
            raise IllegalStateException("Already connected to server.")

    def check_not_connected(self):
        if self.ctx is None:
            raise IllegalStateException("Not connected to server.")
        if self.ctx.isTerminated():
            raise IllegalStateException("Session is terminated.")

    def get_command_context(self):
        return self.ctx

    @staticmethod
    def _construct_uri(protocol, host, port):
        try:
            uri = URI(protocol, None, host, port, None, None, None)
            return uri.toString().substring(2) if protocol is None else uri.toString()
        except URISyntaxException as e:
            raise IllegalStateException("Unable to construct URI.", e)

    def embedded(self, jboss_home, server_config=None):
        self.check_already_connected()
        try:
            self.ctx = CommandContextFactory.getInstance().newCommandContext()
            # TODO factor this out into the instantiation
            # apparently needed in the embedded cli for module loading
            System.setProperty('jboss.home.dir', jboss_home)
            # --std-out=echo
            self.ctx.handle(
                "embed-server --jboss-home=" + jboss_home + " --server-config=" + server_config)
        except CliInitializationException as e:
            raise IllegalStateException("Unable to initialize command context.", e)
        except CommandLineException as ce:
            raise IllegalStateException("Unable to connect to controller.", ce)

    def connect(self):
        self.check_already_connected()
        try:
            self.ctx = CommandContextFactory.getInstance().newCommandContext()
            self.ctx.connectController()
        except CliInitializationException as ie:
            raise IllegalStateException("Unable to initialize command context.", ie)
        except CommandLineException as ce:
            raise IllegalStateException("Unable to connect to controller.", ce)

    def connect(self, protocol=None, controller_host=None, controller_port=None, username=None, password=None,
                client_bind_address=None):
        self.check_already_connected()
        try:

            builder = CommandContextConfiguration.Builder()

            if controller_host is not None and controller_port is not None:
                builder.setController(self._construct_uri(protocol, controller_host, controller_port))

            if controller_port is not None:
                builder.setController(controller_port)

            if username is not None:
                builder.setUsername(username)

            if password is not None:
                builder.setPassword(password)

            if client_bind_address is not None:
                builder.setClientBindAddress(client_bind_address)

            self.ctx = CommandContextFactory.getInstance().newCommandContext(builder.build())
            self.ctx.connectController()

        except CliInitializationException as ie:
            raise IllegalStateException("Unable to initialize command context.", ie)
        except CommandLineException as ce:
            raise IllegalStateException("Unable to connect to controller.", ce)

    def disconnect(self):
        try:
            self.check_not_connected()
            self.ctx.terminateSession()
        finally:
            self.ctx = None

    def cmd(self, cli_command):
        self.check_not_connected()
        try:
            request = self.ctx.buildRequest(cli_command)
            response = self.ctx.getModelControllerClient().execute(request)
            return Result(cli_command, request=request, response=response)
        except CommandFormatException as cfe:
            try:
                self.ctx.handle(cli_command)
                return Result(cli_command, exit_code=self.ctx.getExitCode())
            except CommandLineException as cle:
                raise CommandError("Error handling command: %s" % cli_command, cle)
        except IOException as ioe:
            raise IllegalStateException("Unable to send command " + cli_command + " to server.", ioe)

    def batch_start(self):
        self.check_not_connected()
        try:
            bm = self.ctx.getBatchManager()
            if not bm.isBatchActive():
                bm.activateNewBatch()
        except Exception as e:
            raise IllegalStateException("Failed to start batch.", e)

    def batch_reset(self):
        self.check_not_connected()
        try:
            bm = self.ctx.getBatchManager()
            if bm.isBatchActive():
                bm.discardActiveBatch()

        except Exception as e:
            raise IllegalStateException("Failed to reset batch.", e)

    def batch_is_active(self):
        self.check_not_connected()
        try:
            bm = self.ctx.getBatchManager()
            return bm.isBatchActive()
        except Exception as e:
            raise IllegalStateException("Failed to test if batch is active.", e)

    def batch_add_cmd(self, batch_command):
        self.check_not_connected()
        try:
            cmd = self.ctx.toBatchedCommand(batch_command)
            bm = self.ctx.getBatchManager()

            batch = bm.getActiveBatch()
            batch.add(cmd)

        except Exception as e:
            raise IllegalStateException("Failed to add %s to batch command." % batch_command, e)


class Result(object):
    def __init__(self, cli_command=None, request=None, response=None, exit_code=None):
        self.cliCommand = cli_command
        self.request = request
        self.response = response
        if exit_code is None:
            self.is_success = response is not None and response.get("outcome").asString() == u'success'
            self.is_local_command = False
        else:
            self.is_success = True if exit_code == 0 else False
            self.is_local_command = True

    def getRequest(self):
        return self.request

    def getResponse(self):
        return self.response

    def isSuccess(self):
        return self.is_success



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ParameterError, CommandError
from jyboss.command.core import BaseJBossModule
from jyboss.logging import debug

from java.io import File

__metaclass__ = type

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()


class ModuleModule(BaseJBossModule):
    MODULE_PARAMS = [
        {'name': 'name', 'type': 'str'},
        {'name': 'property', 'type': 'property'},
        {'name': 'enabled', 'type': 'list'}
    ]

    def __init__(self, context=None):
        super(ModuleModule, self).__init__(path='module ', context=context)

    def apply(self, module=None, **kwargs):

        modules = self._format_apply_param(module)

        changes = []

        for mod in modules:
            state = self._get_param(mod, 'state')

            if state not in ['present', 'absent']:
                raise ParameterError('The module state is not one of [present|absent]')

            if state == 'present':
                args = ['%s add' % self.path]
            elif state == 'absent':
                args = ['%s remove' % self.path]

            delim = mod.get('resource-delimiter', File.pathSeparatorChar)
            module_base = mod.get('module-base', 'system.layers.base')

            for k, v in mod.items():
                if k in ['name']:
                    args.append('--%s=%s.%s' % (k, module_base, v))
                elif k in ['slot', 'main-class', 'module-xml']:
                    args.append('--%s=%s' % (k, v))
                elif k in ['resources']:
                    if isinstance(v, list):
                        args.append('--%s=%s' % (k, delim.join(v)))
                    else:
                        args.append('--%s=%s' % (k, v))
                elif k in ['properties']:
                    if isinstance(v, dict):
                        args.append('--%s=%s' % (k, ','.join('%s=%s' % (name, value) for name, value in iteritems(v))))
                    else:
                        raise ParameterError('Format of %s parameter is invalid.' % k)

            try:
                cmd = ' '.join(args)
                debug('module: execute command %s' % cmd)
                self.cmd(cmd)
                changes.append(
                    {'module': self._get_param(mod, 'name'), 'action': 'add' if state == 'present' else 'delete'})
            except CommandError as ce:
                debug('module command error %r' % ce)
                self.debug_errm(ce)
                cause = getattr(ce, 'cause') if hasattr(ce, 'cause') else None
                msg = getattr(cause, 'message') if cause is not None and hasattr(cause, 'message') else ''
                if msg.find('already exists') != -1 or msg.find('Failed to locate module') != -1:
                    # ignore as nothing to do
                    debug('Ignoring module deploy error: %s' % msg)
                else:
                    raise ce
            except Exception as coe:
                debug('module other error %r' % coe)
                self.debug_errm(coe)
                raise coe

        return changes if len(changes) > 0 else None

    def debug_errm(self, err, level=0):
        msg = err.message if 'message' in err and err['message'] is not None else ''
        debug(msg.rjust(level + 1, '>'))
        if 'cause' in err and err['cause'] is not None:
            self.debug_errm(err.cause, level + 1)



#!/usr/bin/env jython

#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements. See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership. The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.
#

try:
    from setuptools import setup, Extension
except:
    from distutils.core import setup, Extension

setup(
    name='jyboss',

    version='0.2.8',

    url='https://github.com/fareliner/jyboss-cli',

    author='Niels Bertram',
    author_email='nielsbne@gmail.com',

    description='Jython CLI for JBoss Application Server',

    license='Apache License 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: System :: Software Distribution',
        'Topic :: System :: Systems Administration',

        # released under Apache 2 License
        'License :: OSI Approved :: Apache Software License',

        # the language used by the author
        'Natural Language :: English',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',

        # only works on Jython
        'Programming Language :: Python :: Implementation :: Jython'
    ],

    keywords='jboss wildfly development cli',

    # only have 2 dependencies
    install_requires=[
        'simplejson',
        'PyYAML'
    ],

    # prepare for testing with nose
    test_suite='nose.collector',
    tests_require=[
        'nose',
        'pathlib2'
    ],

    # manually define packages
    py_modules=[
        'jyboss.context',
        'jyboss.cli',
        'jyboss.exceptions',
        'jyboss.logging',
        'jyboss.command.core',
        'jyboss.command.undertow',
        'jyboss.command.extension',
        'jyboss.command.security',
        'jyboss.command.keycloak',
        'jyboss.command.ee',
        'jyboss.command.weld',
        'jyboss.command.modcluster',
        'jyboss.command.datasources',
        'jyboss.command.module',
        'jyboss.command.deployment',
        'jyboss.command.jgroups',
        'jyboss.command.infinispan',
        'jyboss.command.interface',
        'jyboss.command.binding',
        'jyboss.ansible'
    ]

)



# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)

from jyboss.exceptions import ParameterError, NotFoundError
from jyboss.command.core import BaseJBossModule

__metaclass__ = type

try:
    dict.iteritems
except AttributeError:
    # Python 3
    def iteritems(d):
        return d.items()
else:
    # Python 2
    def iteritems(d):
        return d.iteritems()

try:
    import simplejson as json
except ImportError:
    import json


class InterfaceModule(BaseJBossModule):
    INTERFACE_PARAMS = [
        'any-address',
        'inet-address',
        'loopback-address',
        'multicast',
        'nic',
        'point-to-point',
        'site-local-address',
        'up',
        'any',
        'link-local-address',
        'loopback',
        'nic-match',
        'not',
        'public-address',
        'subnet-match',
        'virtual'
    ]

    def __init__(self, context=None):
        super(InterfaceModule, self).__init__(path='/interface=%s', context=context)

    def apply(self, interface=None, **kwargs):

        interfaces = self._format_apply_param(interface)

        changes = []

        for iface in interfaces:
            state = self._get_param(iface, 'state')
            if state == 'present':
                changes += self.apply_present(iface)
            elif state == 'absent':
                changes += self.apply_absent(iface)
            else:
                raise ParameterError('The interface state is not one of [present|absent]')

        return changes if len(changes) > 0 else None

    def apply_present(self, interface):
        name = self._get_param(interface, 'name')
        iface_path = self.path % name
        changes = []

        try:
            iface_dmr = self.read_resource_dmr(iface_path, recursive=True)
            fc = dict(
                (k, v) for (k, v) in iteritems(interface) if k in self.INTERFACE_PARAMS)
            a_changes = self._sync_attributes(parent_node=iface_dmr,
                                              parent_path=iface_path,
                                              target_state=fc,
                                              allowable_attributes=self.INTERFACE_PARAMS)
            if len(a_changes) > 0:
                changes.append({'interface': name, 'action': 'update', 'changes': a_changes})
        except NotFoundError:
            iface_params = self.convert_to_dmr_params(interface, self.INTERFACE_PARAMS)
            self.cmd('%s:add(%s)' % (iface_path, iface_params))
            changes.append({'interface': name, 'action': 'add'})
        return changes

    def apply_absent(self, interface):
        name = self._get_param(interface, 'name')
        iface_path = self.path % name
        changes = []

        try:
            self.read_resource_dmr(iface_path)
            self.cmd('%s:remove()' % iface_path)
            changes.append({'interface': name, 'action': 'delete'})
        except NotFoundError:
            pass

        return changes
